from pygmo import problem, population, algorithm, sga
from PyCoptimizer.framework.comp_clean import CompClean
from PyCoptimizer.results import Results


class Island(object):
    """ Class to run the optimizer using the island method 
        The island method is provided by the pygmo module
    """

    def __init__(self, user_class, population):
        """ Constructor
        Parameters
        ----------
        user_class: class
            Class provided by the user
        dim: value
            It represents the number of flags to be used, 
            the dimension of the problem
        compClean: class
            Class to compile the code and clean the executable files
        no_pop: value
            Value containing the population size
        no_generations: value
            Value containing the number of generations
        no_individuals: value
            Value containing the individual size (number of genes)
        crossover_rate: value
            Crossover rate to be used in the genetic algorithm
        mutation_rate: value
            Mutation rate to be used in the genetic algorithm
        population: list
            List containing the inicial flags to be used
        prob: class
            Problem defined by pygmo
        pop: class
            Population defined by pygmo
        """
        self.user_class = user_class
        self.dim = len(self.user_class.flags_list)
        self.compClean = CompClean(self.user_class)
        self.no_pop = 0
        self.no_generations = 0
        self.no_individuals = 0
        self.crossover_rate = 0
        self.mutation_rate = 0
        self.population = population
        self.prob = None
        self.pop = None

    def defining_values(self):
        """ Method to set the class parameters defined 
            in the constructor
        """
        self.no_pop = self.user_class.dict_optimization["no_pop"]
        self.no_generations = self.user_class.\
            dict_optimization["no_generations"]
        self.no_individuals = self.user_class.\
            dict_optimization["individual_size"]
        self.crossover_rate = self.user_class.\
            dict_optimization["crossover_rate"]
        self.mutation_rate = self.user_class.\
            dict_optimization["mutation_rate"]
        self.prob = problem(self)
        self.pop = self.create_population(self.prob, self.population)

    def fitness(self, x):
        """ Method to calculate the fitness value
        This method uses methods defined by the user
        The method name and the return is required for the 
        pygmo structure to work the island method

        Returns
        -------
        y : float
            This method returns the fitness value.
        """
        self.user_class.pre()
        self.compClean.compile_code(x)
        y = self.user_class.evaluation_function()
        self.compClean.clean_code()
        self.user_class.pos()
        return [self.user_class.min_max*y]

    def get_bounds(self):
        """ Method returns the bound values
        The method name and the return is required for the 
        pygmo structure to work the island method
        Returns
        -------
        bounds : tuple
            This method returns the bounds of the problem.
        """
        return ([0] * self.dim, [1] * self.dim)

    def create_population(self, prob, population_list):
        """ Method to set the population using push_back method

        Returns
        -------
        pop : Population instance
            This method returns an instance of Population class.
        """
        pop = population(prob)
        for i in range(0, self.no_pop):
            pop.push_back(x=population_list[i])
        return pop

    def island_method(self, population, flag_option, user_class):
        """ Method to run the optimizer
        In this optimizer, it will be used the genetic algorithm, 
        provided by pygmo module
        This method will return a tuple with the generation 
        and their correspondent best evaluation

        Returns
        -------
        log : tuple
            This method returns a tuple containg 
            the generations and their corresponding est evaluations.
        """
        self.defining_values()
        algo = algorithm(sga(gen=self.no_generations,
                             cr=self.crossover_rate,
                             m=self.mutation_rate,
                             selection="tournament"))


        algo.set_verbosity(2)
        self.pop = algo.evolve(self.pop)
        logs = algo.extract(sga).get_log()
        return ([gen[0] for gen in logs], [val[2] for val in logs],
                self.pop.champion_x)


def main(user_class, population, flag_option, results):
    """ Main method to call the island method
    """
    island = Island(user_class, population)
    generations, values, best_ind = island.island_method(
        population, flag_option, user_class)
    results.get_results(generations, values, best_ind)


if __name__ == "__main__":
    main()



from abc import ABCMeta, abstractmethod
from os import getcwd


class UserParameters(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def __init__(self):
        """Method to be implemented by the user
            Set the initial list of flags to be used in the 
            optimization
            Set the path to compile
            Set if the user wants to minize or maximize the problem
        Parameters
        ----------
        flags_list: list
            The list with initial flags
            The default list is composed with flags for gcc compiler
        compile_path: string
            Path containing where is the file to compile
        compile_command: string
            Command indicating how to compile the code
            The default command is "make"
        clean_path: string
            Path containing where is the file to clean the executable 
            files
        clean_command: string
            Command indicating how to clean the executable files
            The default command is "make clean"
        min_max: value
            A value representing if the user wants to minimize or 
            maximize the problem
            For minimize, the value is 1 and maximize, -1
            The default is minimize
        macro: string
            String to indicate to the compiler the flags to be used.
        method_optimization: string
            String containing the name of the method to do 
            the optimization.
            The string can be: island.
        dict_optimization: dictionary
            This dictionary needs to have all information for the 
            optimization method. 
            For the island method, you must give these informations:
                -no_generations: number of generations
                -no_pop: population size
                -individual_size: individual size; this size must be 
                equal than the number of flags in the 
                flags_list
                -crossover_rate: crossover rate; must be between 0 
                and 1
                -mutation_rate: mutation rate; must be between 0 and 1
        static_flags: string
            String containing flags that will be used in every compilation.    
        """
        self.flags_list = ["-O1", "-O2", "-O3"]
        self.compile_path = getcwd()
        self.compile_command = "make"
        self.clean_path = getcwd()
        self.clean_command = "make clean"
        self.min_max = 1
        self.macro = "CXXFLAGS"
        self.method_optimization = "island"
        self.dict_optimization = {"no_generations": 3,
                                  "no_pop": 2,
                                  "individual_size": 3,
                                  "crossover_rate": 0.3,
                                  "mutation_rate": 0.1}
        self.static_flags = ""
        return

    @abstractmethod
    def pre(self):
        """Method to be implemented by the user
           This method will be executed before every compilation and evaluation.
        """
        return

    @abstractmethod
    def arguments_to_run_code(self):
        """Method to be implemented by the user
        Method to have the way to run the executable file 

        Returns
        -------
        command: string
        Command indicating how to run the executable file.

        """
        return

    @abstractmethod
    def evaluation_function(self, x):
        """Method to be implemented by the user
            Method to do the evaluation of the code to be optimized

        Returns
        -------
        value : float
            This method returns the value to be optimized.
        """
        return

    @abstractmethod
    def pos(self):
        """Method to be implemented by the user
           This method will be executed after cleaning the executable files.
        """
        return



from os import getcwd, chdir, environ
from subprocess import Popen, PIPE


class CompClean(object):
    """ Class to compile the code and clean the executable files
    """

    def __init__(self, user_class):
        self.user_class = user_class

    def __change_flags(self, individual, options, no_individuals):
        """ Method returns the population with the flags
        In this method, if the bool value is 1, 
        it is switched to the correspondent flag

        Returns
        -------
        individual_def : list
            List containing the chosen flags to be 
            used in the compilation.
        """
        individual_def = ""
        for ind in range(0, len(individual)):
            if individual[ind] > 0.5:
                individual_def = individual_def + " " + options[ind]
        individual_def = individual_def + " " + self.user_class.static_flags
        individual_def = '"' + individual_def + '"'
        return individual_def

    def compile_code(self, individual):
        """ Method to compile the target code
            This method uses methods defined by the user  
        """
        compile_path = self.user_class.compile_path
        compile_command = self.user_class.compile_command
        flags = self.user_class.flags_list
        no_individuals = self.user_class.dict_optimization["individual_size"]
        individual = self.__change_flags(individual, flags, no_individuals)
        compile_command = "{}={} {}".format(
            self.user_class.macro, individual, compile_command)

        path = getcwd()
        chdir(compile_path)


        my_env = environ.copy()
        my_env["PATH"] = "/usr/sbin:/sbin:{}".format(my_env["PATH"])

        process = Popen(compile_command, shell=True, stdout=PIPE)
        out, err = process.communicate()
        chdir(path)

    def clean_code(self):
        """ Method to clean the target code
            This method uses methods defined by the user
            It cleans all the executable files
        """
        path = getcwd()
        clean_path = self.user_class.clean_path
        clean_command = self.user_class.clean_command

        chdir(clean_path)

        my_env = environ.copy()
        my_env["PATH"] = "/usr/sbin:/sbin:{}".format(my_env["PATH"])

        process = Popen(clean_command, shell=True, stdout=PIPE)
        out, err = process.communicate()
        chdir(path)



from importlib import import_module
from importlib import import_module
from inspect import getmembers, isclass
import random
import sys
import os
from PyCoptimizer.framework.island import main
from PyCoptimizer.results import Results


class COptimizer(object):
    """ Class to run the optimizer 
    """

    def __init__(self):
        """ Constructor
        Parameters
        ----------
        user_class: class
            Class defined by the user and get in the command line
        results: class
            Class that will show the results at the end
        """
        sys.path.append(os.path.dirname(sys.argv[0]))
        module_name = os.path.splitext(os.path.basename(sys.argv[0]))[0]
        module_user = import_module(module_name)
        sys.path.pop()
        if len(sys.argv) == 2:
            result_path = './'
        else:
            result_path = sys.argv[2]

        result_type = sys.argv[1]
        klasses = getmembers(module_user, isclass)
        my_class = getattr(module_user, klasses[1][0])
        self.user_class = my_class()
        self.results = Results(self.user_class, result_type, result_path)

    def get_flags(self):
        """ Getting the flag options defined by the user

        Returns
        -------
        flags_list: list
            List containing the flags defined by the user
        """
        return self.user_class.flags_list

    def construct_first_population(self):
        """ Constructing the first population 
        This first population is composed by 0's and 1's, 
        representing if the flag is present in the population 
        or not

        Returns
        -------
        individuals: list
            List containing the actived flags in the population
        """
        options = self.get_flags()
        no_pop = self.user_class.dict_optimization["no_pop"]
        individuals = []
        for ind in range(0, no_pop):
            individual_first = []
            for individual in range(0, len(options)):
                individual_first.append(random.randint(0, 1))
            individuals.append(individual_first)
        return individuals

    def run_optimization(self):
        """ Run the optimizer using the informations above
        """
        population = self.construct_first_population()
        flags = self.get_flags()
        main(self.user_class, population, flags, self.results)

    def main(self):
        self.run_optimization()



from decimal import Decimal
from math import fmod
import matplotlib
matplotlib.use('Agg')
import matplotlib.image as mpimg
import matplotlib.pyplot as plt


class Results(object):

    def __init__(self, user_class, result_type, path='./'):
        """ Constructor

        Parameters
        ----------
        user_class: class defined by the user
        result_type: resolve if the result will be printed in the screen or
                    will be in the graphic
        path: path where to save the results 
        """
        self.user_class = user_class
        self.no_pop = user_class.dict_optimization['no_pop']
        self.path = path
        self.result_type = result_type
        self.best_ind = []
        self.best_ind_bin = []

    def translate_individual(self):
        if len(self.best_ind_bin):
            list_ind = zip(self.best_ind_bin, self.user_class.flags_list)
            for individual in list_ind:
                if individual[0]:
                    self.best_ind.append(individual[1])

    def log(self, list_generations, list_values):
        """ Method to print in the screen the main results
        """
        print("Results:\n")
        print('{} {:^10}'.format('Generation: ', 'Value: '))
        for res in range(len(list_generations)):
            print('{} {:^20}'.format(list_generations[res], list_values[res]))
        print("Best individual: ", self.best_ind)
        print("Static flags: ", self.user_class.static_flags)

    def graphic(self, list_generations, list_values):
        """ Method to draw in a graphic the main results
        """
        plt.plot(list_generations, list_values, linewidth=2.5,
                 label="Island: " + str('%.2E' % Decimal(list_values[-1])))
        plt.legend(bbox_to_anchor=(0.5, 0), loc=3,
                   ncol=2, mode="expand", borderaxespad=0.)
        plt.xlabel('generations')
        plt.ylabel('best evaluate')
        plt.title('Generation with: population= ' + str(self.no_pop),
                  bbox={'facecolor': '0.8', 'pad': 5})
        plt.savefig("{}island_{}.png".format(self.path, str(self.no_pop)))
        print("Best individual: ", self.best_ind)
        print("Static flags: ", self.user_class.static_flags)

    def get_results(self, list_generations, list_values, best_ind):
        """ Method to choose how the user wants to see their results
        """
        self.best_ind_bin = best_ind
        self.translate_individual()
        if self.result_type == 'log':
            self.log(list_generations, list_values)
        else:
            self.graphic(list_generations, list_values)



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='PyCoptimizer',
    version='0.1.2',
    description='Python Code Optimization using compilation flags',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='Leticia Figueiredo',
    author_email='letfreirefigueiredo@gmail.com',
    packages=find_packages(),
    license='Apache License',
    install_requires=['setuptools', 'numpy', 'matplotlib'],
    package_dir={'COptimizer': './PyCoptimizer/coptimizer',
                 'UserParameters': './PyCoptimizer/user_parameters',
                 'Results': './PyCoptimizer/results',
                 'CompClean': './PyCoptimizer/framework/comp_clean',
                 'Island': './PyCoptimizer/framework/island'}
)
